/******************************************************************************/
/*                                                                            */
/*    Copyright (c) Centre de traitement en imagerie virtuelle inc.           */
/*    All Rights Reserved.                                                    */
/*                                                                            */
/*    These coded instructions, statements, and computer programs  contain    */
/*    unpublished  proprietary and  confidential information of  Centre de    */
/*    traitement  en  imagerie virtuelle inc. (CTIV), and are protected by    */
/*    Federal copyright law. They may not be disclosed to third parties or    */
/*    or duplicated in any form, in whole or in  part, without  the  prior    */
/*    written consent of CTIV.                                                */
/*                                                                            */
/******************************************************************************/

// ===========================================================================
// = LIBRARY
//     3delight
// = FILENAME
//     dlRiOptions.h
// = AUTHOR(S)
//     Patrick Fournier
// = VERSION
//     $Revision: 1.12 $
// = DATE RELEASED
//     $Date: 2000/06/20 19:05:26 $
// = RCSID
//     $Id: dlRiOptions.h,v 1.12 2000/06/20 19:05:26 marco Exp $
// = COPYRIGHT
//     Copyright  1999-2000 Centre de traitement en imagerie virtuelle inc.
//     All Rights Reserved.
// ===========================================================================

#ifndef __dlRIOptions_H
#define __dlRIOptions_H

#include "dlToken.h"
#include "dlFilterFunction.h"
#include "dlDataFormat.h"

#ifdef WIN32
#include "3DeLightSTL.h"
#else
#include "vector.h"
#endif

#ifndef __declspec
#ifndef WIN32
#define __declspec(x)
#endif
#endif


// ===========================================================================
//      C L A S S E S
// ===========================================================================
class DlRenderingSession;
class DlDspyServer;
class DlSpaceRegistry;

// ===========================================================================
//
//      CLASS DlRIOptions
//
// ===========================================================================

/**
  DlRIOptions.

  \begin{description}

  \item[CLASS TYPE]
  Used to hold and set rendering options.

  \item[DESCRIPTION]
  --

  \item[NOTES]
  --

  \end{description}

  @author Patrick Fournier
  @see --
  */
class DlRIOptions
{
public:
	typedef enum {
		e_hiderHidden = 0
	} DlHiderType;
    
	typedef enum {
		e_regularGrid = 0,
		e_stochasticJitter,
		e_monteCarlo
	} DlSamplingMode;

	typedef enum {
		e_horizontal = 0,
		e_vertical,
		e_spiral,
        e_random
	} DlBucketOrder;

	DlRIOptions(DlRenderingSession* i_session, DlSpaceRegistry* i_spaceRegistry);
	DlRIOptions(const DlRIOptions& i_options);
	~DlRIOptions(void);
    
	// Camera
	void Format(int i_xRes, int i_yRes, float i_pixelAspectRatio, DlSpaceRegistry* i_spaceRegistry);
	void FrameAspectRatio(float i_ratio, DlSpaceRegistry* i_spaceRegistry);
	void ScreenWindow(float i_left, float i_right, float i_bottom, float i_top, DlSpaceRegistry* i_spaceRegistry);
	void CropWindow(float i_xmin, float i_xmax, float i_ymin, float i_ymax);
	void NullProjection(DlSpaceRegistry* i_spaceRegistry);
	void PerspectiveProjection(float i_fov, DlSpaceRegistry* i_spaceRegistry);
	void OrthographicProjection(DlSpaceRegistry* i_spaceRegistry);
	void Clipping(float i_near, float i_far, DlSpaceRegistry* i_spaceRegistry);
	void DepthOfField(float i_fstop, float i_focalLength, float i_focalDistance);
	void Shutter(float i_min, float i_max);

	// Display
	void PixelSamples(float i_xsamples, float i_ysamples);
	void PixelFilter(DlFilterFunction i_filterFunc, float i_xwidth, float i_ywidth);
	void Exposure(float i_gain, float i_gamma);
	void ColorQuantize(float i_one, float i_min, float i_max, float i_dither);
	void DepthQuantize(float i_one, float i_min, float i_max, float i_dither);
	void Display(const char* i_name, const char* i_displayDriverPath, const char* i_type, DlDataFormat i_mode, 
		int n, char const* tokens[], void const* parms[], bool i_addingDisplay,
				 DlSpaceRegistry* i_spaceRegistry);

	// Additional Options
	void Hider(const char* i_type, int n, const char* tokens[], void* parms[]);

	// Implementation specific
	void LimitsBucketSize(unsigned int i_sizex, unsigned int i_sizey);
	void LimitsGridSize(unsigned int i_gridSize);
	void LimitsTextureMemory(unsigned int i_size);
	void LimitsZThreshold(float i_threshold);
	void LimitsExtremeDisplacement(unsigned int i_limit);
	void LimitsEyeSplits(unsigned int i_eyeSplits);

	void RenderBucketOrder( const char *i_bucketOrder );
			
	void ShadowBias0(float i_bias);
	void ShadowBias1(float i_bias);
	void ShadowSample(int i_sample);

	void SearchPathShader(char* i_path);
	void SearchPathTexture(char* i_path);
	void SearchPathDisplay(char* i_path);
	void SearchPathArchive(char* i_path);
	
	void SetAllOptions(DlSpaceRegistry* i_spaceRegistry);

	DlHiderType HiderType(void) const { return m_hiderType; }

	static void FreeDisplays(void);

	// Check if option parameter is valid and return the number and it's type
	//
	__declspec(dllexport) static bool CheckIfOptionValid(const char          *i_option, 
														 const char          *i_param,
														 int                 &o_nbElems, 
														 DlTokenInfo::DlType &o_type);

	/////////////////////////////////////////////////////////////////////////
	// Access functions (for shaders)
	/////////////////////////////////////////////////////////////////////////

	__declspec(dllexport) float NearClippingPlane(void) const
        {
            return m_nearClipPlane;
        }

	__declspec(dllexport) float FarClippingPlane(void) const
        {
            return m_farClipPlane;
        }
	

protected:
	void SetRasterSpace(DlSpaceRegistry* i_spaceRegistry) const;
	void SetScreenSpace(DlSpaceRegistry* i_spaceRegistry) const;
	
	void SetFormat(DlSpaceRegistry* i_spaceRegistry) const;
	void SetScreenWindow(DlSpaceRegistry* i_spaceRegistry) const;
	void SetCropWindow() const;
	void SetProjection(DlSpaceRegistry* i_spaceRegistry) const;
	void SetClipping(DlSpaceRegistry* i_spaceRegistry) const;
	void SetDepthOfField() const;
	void SetShutter() const;

	void SetPixelSamples() const;
	void SetPixelFilter() const;
	void SetExposure() const;
	void SetColorQuantize() const;
	void SetDepthQuantize() const;
	void SetDisplay() const;

	void SetHider() const;

	void SetLimitsBucketSize() const;
	void SetLimitsGridSize() const;
	void SetLimitsTextureMemory() const;
	void SetLimitsZThreshold() const;
	void SetLimitsExtremeDisplacement() const;
	void SetLimitsEyeSplits() const;

	void SetRenderBucketOrder();

	void SetShadowBias0() const;
	void SetShadowBias1() const;
	void SetShadowSample() const;

	void SetSearchPathShader() const;
	void SetSearchPathTexture() const;
	void SetSearchPathDisplay() const;
	void SetSearchPathArchive() const;
	
	typedef enum {
		e_nullProjection,
		e_orthographicProjection,
		e_perspectiveProjection
	} DlProjectionType;

	class DlDisplayDesc
	{
	public:
		DlDspyServer* m_display;
		DlDataFormat m_mode;
	};

	DlRenderingSession *m_session;
	bool m_screenWindowCalled;
	bool m_formatCalled;

	int m_xres, m_yres;
	float m_pixelAspectRatio;

	float m_swLeft, m_swRight, m_swBottom, m_swTop;
	
	float m_cropXMin, m_cropXMax, m_cropYMin, m_cropYMax;
	
	DlProjectionType m_projectionType;
	float m_fov;
	
	float m_nearClipPlane, m_farClipPlane;
	
	float m_fstop, m_focalLenght, m_focalDistance;
	
	float m_shutterOpen, m_shutterClose;

	float m_xsamples, m_ysamples;
	
	DlFilterFunction m_filterFunc;
	float m_filterWidth, m_filterheight;
	
	float m_gain, m_gamma;

	float m_colorOne, m_colorMin, m_colorMax, m_colorDitherAmplitude;

	float m_depthOne, m_depthMin, m_depthMax, m_depthDitherAmplitude;

	vector<DlDisplayDesc*> m_displays;

	DlHiderType m_hiderType;
	DlSamplingMode m_samplingMode;

	// int m_numColorSamples;
	// float *m_toRGB;
	// float *m_fromRGB;

	// float m_relativeDetail;

	unsigned int m_bucketSizeX, m_bucketSizeY;
	unsigned int m_gridSize;
	unsigned int m_textureMemory;
	float m_zThreshold;
	unsigned int m_extremeDisplacement;

	float m_shadowBias0;
	float m_shadowBias1;
	int m_shadowSample;

	char* m_shaderPaths;
	char* m_texturePaths;
	char* m_displayPaths;
	char* m_archivePaths;

	DlBucketOrder m_bucketorder;

	static vector<DlDspyServer*> m_displayList; // used for deleting all displays

	struct OptionParamInfo 
	{
		const char          *m_optionName;
		const char          *m_paramName;
		DlTokenInfo::DlType  m_type;
		int                  m_nbElems;
	};

	static OptionParamInfo sm_optionsInfo[];
};

#endif
