/******************************************************************************/
/*                                                                            */
/*    Copyright (c)The 3Delight Developers.                                   */
/*    All Rights Reserved.                                                    */
/*                                                                            */
/******************************************************************************/


// ===========================================================================
// = LIBRARY
//     3delight
// = FILENAME
//     dlTokenInfo.h
// = AUTHOR(S)
//     Patrick Fournier
// = VERSION
//     $Revision: 1.21 $
// = DATE RELEASED
//     $Date: 2003/05/13 17:38:52 $
// = RCSID
//     $Id: dlTokenInfo.h,v 1.21 2003/05/13 17:38:52 olivier Exp $
// ===========================================================================


#ifndef __dlTokenInfo_HH
#define __dlTokenInfo_HH

#include "dlTokenRIGlue.h"

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include "dlInterface.h"

#define NUM_TYPES 12 

static const int k_typeChannels[NUM_TYPES] = { 0, // bad
                              1, // float
                              3, // point
                              3, // color
                              1, // string
                              1, // int
                              1, // pointer
                              16, // matrix
                              3, // vector
                              3, // normal
                              1, // bool
							  4 /*hpoint*/ };

const int k_typeSizes[NUM_TYPES] = { 0,
                              sizeof(RIGFloat),
                              sizeof(RIGPoint),
                              sizeof(RIGColor),
                              sizeof(RIGString),
                              sizeof(RIGInt),
                              sizeof(RIGPointer),
                              sizeof(RIGMatrix),
                              sizeof(RIGVector),
                              sizeof(RIGNormal),
                              sizeof(RIGBoolean),
							  sizeof(RIGHpoint) };
  
class DlTokenInfo
{
public:
    typedef enum
    {
        e_badClass,
        e_constant,
        e_uniform,
        e_varying,
        e_vertex,
		e_facevarying,
		e_facevertex
    } DlClass;

    typedef enum
    {
        e_badType = 0,
        e_RtFloat = 1,
        e_RtPoint = 2,
        e_RtColor = 3,
        e_RtString = 4,
        e_RtInt = 5,
        e_RtPtr = 6,
        e_RtMatrix = 7,
        e_RtVector = 8,
        e_RtNormal = 9,
        e_RtBool = 10,
		e_RtHpoint = 11
    } DlType;

    DlTokenInfo() : m_class(e_badClass), m_type(e_badType), m_arraySize(0) {}
    DlTokenInfo(const DlTokenInfo& i_info)
        : m_class(i_info.m_class),
          m_type(i_info.m_type),
          m_arraySize(i_info.m_arraySize)
        {
            if (((m_class == e_varying) && (m_type == e_RtString)) ||
                ((m_class == e_vertex) && (m_type == e_RtString)))
            {
                m_class = e_badClass;
            }
        }

	/*
		Class
		
		returns the class of this variable
		
		NOTE
		can be "uniform", "varying", "constant" ...
	*/
	DlClass Class(void) const { return m_class; }
	
	/*
		Type
		
		returns the type of this variable

		NOTE
		can be "float", "point", "color" ...
	*/
    DlType Type(void) const { return m_type; }

	/*
		IsPointLike (1)

		Returns true if this token is either a point, vector, normal or hpoint
	*/
	bool IsPointLike( void ) const
	{
		return
			m_type==e_RtPoint || m_type==e_RtHpoint || m_type==e_RtNormal ||
				m_type==e_RtVector;
	}

	/*
		IsPointLike (2)
	*/
	static bool IsPointLike( DlType i_type )
	{
		return
			i_type==e_RtPoint || i_type==e_RtHpoint || i_type==e_RtNormal ||
				i_type==e_RtVector;
	}

	/*
		IsTransformable (1)
		
		returns true if this token has to be transform the same way as the
		geometry. 

		NOTE
		- this is true for all "point like" tokens and matrix tokens which
		are "vertex" variables;
		- TODO: It seems that all 'point like variables' have to be transformed,
		and not only 'vertex' variables.
	*/
	bool IsTransformable( void ) const
	{
		return (IsPointLike() || m_type==e_RtMatrix) /*&& m_class==e_vertex*/ ;
	}

	/*
		IsTransformable (2)
	*/
	static bool IsTransformable( DlType i_type, DlClass i_class )
	{
		return (IsPointLike(i_type) || i_type==e_RtMatrix) && i_class==e_vertex;
	}
	
	/*
		NumChannles

		How many channles per single value in this varibale?

		NOTE
		ie : RtVector = 3, RtColor = 3, RtFloat = 1, RtMatrix = 16 ...
	*/
	unsigned NumChannels( void ) const
	{
		return k_typeChannels[m_type];
	}

	static unsigned NumChannels( const DlType i_type )
	{
		return (i_type<NUM_TYPES) ? k_typeChannels[i_type] : 0;
	}

	/*
		ArraySize

		How many array elements does this token contain?

		NOTE
		a token may contain many values if it is an array. if it's not an array,
		this function will always return 1;
	*/
    unsigned ArraySize(void) const { return m_arraySize; }
	
	
	/*
		NumValues
		
		total number of values is defined by : NumChannels * ArraySize
	*/
	unsigned NumValues( void ) const { return NumChannels()*ArraySize(); }

	/*
		SameType

		returns true if this variable is of the same type & class as some
		given token
	*/
	bool SameType( const DlTokenInfo& i_tokenInfo ) const;
	
	/*
	*/
    void Set( DlClass i_class, DlType i_type, unsigned char i_arraySize )
	{
		m_class = i_class;
		m_type = i_type;
		m_arraySize = i_arraySize;

		if( (m_class==e_varying && m_type==e_RtString) ||
		    (m_class==e_vertex && m_type==e_RtString) )
		{
			assert( !"Invalid class/type combination" );
			m_class = e_badClass;
		}
	}

    /** Get a string representing this token's declaration.

        Return a string representing this token's declaration. The string
        is stored as function's static data, so it will be invalid on the
        next GetDeclaration() call.
    */
    const char* GetDeclaration(void) const;

	/*
		GetClass

		Given a string containing some class name (such as "uniform"), this 
		function returns the appropriate DlClass.
	*/
    static DlClass GetClass( const char* i_className );
	DL_INTERFACE static const char *GetClass( const DlClass i_class );

	/*
		GetType

		Given a string containing some type name (such as "color"), this 
		function returns the appropriate DlType... and vis-versa!
	*/
    static DlType GetType( const char* i_typeName );
	DL_INTERFACE static const char *GetType( const DlType i_type );

	/*
		ParseInlineDeclaration

		Given a string containing a variable declaration, such as in
		"uniform color [3]", this procedures returns variable's class
		(e_uniform). type (e_RtColor) and array size (3).
	*/
	DL_INTERFACE static bool ParseInlineDeclaration(
		const char *i_declaration,
		DlClass &o_class,
		DlType &o_type,
		unsigned &o_arraySize,
		char **o_name = 0x0 );

	/*
		SizeOf

		return the size of this token in bytes

		NOTE
		this function uses a precomputed array of sizes.
	*/
    static unsigned SizeOf( DlType i_type, unsigned i_arraySize )
	{
		assert( k_typeSizes[i_type]>0 );
		return k_typeSizes[i_type]*i_arraySize;
	}

    unsigned SizeOf(void) const { return SizeOf(m_type, m_arraySize); }

    /*
		CanAssign

		Returns true if data of type i_info can de assigned to a variable of
		*this type 
	*/
    bool CanAssign( const DlTokenInfo& i_info ) const;

	/*
		ConvertToShaderClass
		
		Converts a RI token class to the equivalent shader token class.
		ie. constant -> uniform, vertex -> varying, facevarying -> varying etc
	*/
	static DlClass ConvertToShaderClass(DlClass i_class);

    /*
		CopyValue

		Copies the value pointed to by i_source at address i_dest.
		In the case of a RtString, also copy the strings.
	
		PARAMETERS
		i_dest    : destination address
		i_source  : source address
		i_type    : type of data to copy
    */
    static void CopyValue(
		void* i_dest,
		const void* i_source,
		const DlTokenInfo& i_token )
	{
		assert( i_dest != 0 );
		assert( i_source != 0 );
		assert( i_token.Type() != e_badType);

		if( i_token.Type() == e_RtString )
		{
			for (unsigned int i = 0; i < i_token.ArraySize(); i++)
			{
                ((char**)i_dest)[i] = strdup(((char**)i_source)[i]);
			}
		}
		else
		{
			memcpy(i_dest, i_source, i_token.SizeOf());
		}
		
		return;
	}
    
    /*
      i_dest: destination address
      i_source: source address
      i_type: type of data to copy
      i_arraySize: number of values in array
      
      Copies the values pointed to by i_source at address i_dest.
      In the case of a RtString, also copy the strings.
    */
    static void CopyArray(
		void* i_dest,
		const void* i_source,
		const DlTokenInfo& i_token,
		unsigned i_arraySize )
	{
		assert(i_dest != 0);
		assert(i_source != 0);
		assert(i_token.Type() != e_badType);
        
		if( i_token.Type() == e_RtString )
		{
			for( unsigned int i = 0; i < i_token.ArraySize() * i_arraySize; i++ )
			{
				((char**)i_dest)[i] = strdup( ((char**)i_source)[i] );
			}
		}
		else
		{
			memcpy( i_dest, i_source, i_token.SizeOf() * i_arraySize );
		}

		return;
	}

	/* Free memory allocated by CopyValue */
    static void FreeValue(
		const void* i_val,
		const DlTokenInfo& i_token)
    {
		if( i_token.Type() != e_RtString )
		{
			return;
		}

		for( unsigned int i = 0; i < i_token.ArraySize(); i++ )
		{
			free(((char**)i_val)[i]);
		}
    }

	/* Free memory allocated by CopyValue */
    static void FreeValue(
		const void* i_val,
		const DlType i_type, 
		unsigned i_arraySize )
    {
		if( i_type != e_RtString )
		{
			return;
		}

		for( unsigned int i = 0; i < i_arraySize; i++ )
		{
			free( ((char**)i_val)[i] );
		}
    }

	/* Free memory allocated by CopyArray */
    static void FreeArray(
		const void* i_array,
		const DlTokenInfo& i_token,
		unsigned int i_arraySize )
	{
		assert( i_array!=0 );
		assert( i_token.Type()!=e_badType );

		if( i_token.Type() != e_RtString )
		{
			return;
		}

		for( unsigned int i = 0; i < i_token.ArraySize() * i_arraySize; i++ )
		{
            if (((char**)i_array)[i])
            {
                free((void*)((char**)i_array)[i]);
            }
		}
	}

    void CheckIntegrity( void ) const
    {
        assert((m_class == e_badClass) ||
               (m_class == e_constant) ||
               (m_class == e_uniform) ||
               (m_class == e_varying) ||
               (m_class == e_vertex));

        assert((m_type == e_badType) ||
               (m_type == e_RtFloat) ||
               (m_type == e_RtPoint) ||
               (m_type == e_RtColor) ||
               (m_type == e_RtString) ||
               (m_type == e_RtInt) ||
               (m_type == e_RtPtr) ||
               (m_type == e_RtMatrix) ||
               (m_type == e_RtBool));
    }
    
protected:
    DlClass m_class;
    DlType m_type;

	/* In an array, the number of elements. Will contain 1 if not an  array */
    unsigned m_arraySize;
};


#endif // __dlTokenInfo_HH

