/******************************************************************************/
/*                                                                            */
/*    Copyright (c)The 3Delight Developers.                                   */
/*    All Rights Reserved.                                                    */
/*                                                                            */
/******************************************************************************/

// ===========================================================================
// = LIBRARY
//     3delight
// = AUTHOR(S)
//     Patrick Fournier
// = VERSION
//     $Revision: 1.28 $
// = DATE RELEASED
//     $Date: 2003/06/27 17:56:40 $
// = RCSID
//     $Id: dlVar.h,v 1.28 2003/06/27 17:56:40 aghiles Exp $
// ===========================================================================

#ifndef __dlVar_H
#define __dlVar_H

#include "dlShaderToken.h"
#include "dlInterface.h"

class DlShapeMesh;

/* Forward declaration (from dlSL.h)
*/
class DlVar;
extern "C" DL_INTERFACE DlVar* DlCreateVarsInit( DlShaderToken *i_st, int i_nb );

/*
	DlVar
	

*/
class DlVar
{
	/* for use of the private method 'InitWithShaderToken' */
	friend class DlEvalIntShader;
	DL_INTERFACE friend DlVar* DlCreateVarsInit( DlShaderToken *, int i_nb );

    /* e_ownValue: set if m_value was allocated by this object */
    typedef enum
    {
		e_clear = 0x0000,
		e_ownValue = 0x0001,
		e_useDefaultValue = 0x0002
    } DlValueFlag;
	
public:

	DlVar();
	DlVar( 
		const DlShaderToken* i_token,
		DlTokenInfo::DlClass i_class,
		unsigned i_numValues,
		const void* i_value );
	DlVar( 
		const DlShaderToken* i_token,
		DlTokenInfo::DlClass i_class,
		DlTokenInfo::DlType i_type,
		unsigned i_numValues,
		const void* i_value );

	DlVar(const DlShaderToken* i_token);
	~DlVar();

	/* Use of this function is prohibited; it is here only
	   for the creation of the visibleVars array in the MSVC C shaders */
	DlVar(const DlVar& i_var);
    
	const DlShaderToken* ShaderToken() const { return m_shaderToken; }

	/* Copy this into v. */
	void CopyTo(DlVar& v) const;
	
	/* Same as copy but do not actually copy the data, only the
	   pointer */
	void LinkTo(DlVar &v) const;

	/* 
		CopyValue

		Allocate space and copy the array i_value in it.
		
		NOTE
		- DlVar does not keep a copy of the space name string, but just a
		pointer to it.
     */
	void CopyValue(
		const char* i_spaceName,
		DlTokenInfo::DlClass i_valueClass,
		DlTokenInfo::DlType i_type,
		unsigned i_numValues, const void* i_value);

	void CopyValue(
		const char* i_spaceName,
		DlTokenInfo::DlClass i_valueClass,
		unsigned i_numValues, const void* i_value)
	{
		CopyValue(
			i_spaceName, i_valueClass,
			DlTokenInfo::e_badType, i_numValues, i_value );
	}
    
	/*
		SetValue
		
		Just sets m_value = i_value (no allocation, no copy)
       
		NOTE
		- DlVar does not keep a copy of the space name string, but just
		a pointer to it.
    */
	void SetValue(
		const char* i_spaceName,
		DlTokenInfo::DlClass i_valueClass,
		DlTokenInfo::DlType i_type,
		unsigned i_numValues, void* i_value );

	void SetValue(
		const char* i_spaceName,
		DlTokenInfo::DlClass i_valueClass,
		unsigned i_numValues, void* i_value )
	{
		SetValue(
			i_spaceName, i_valueClass,
			DlTokenInfo::e_badType, i_numValues, i_value );
	}

	/*
		ClearValue

		Clears the content of this variable
	*/
    void ClearValue( void );
    
	/*
		ValueClass
		
		returns the class of this variable
	*/
	DlTokenInfo::DlClass ValueClass() const { return m_valueClass; }

	/*
		Value
		
		returns a ptr to the value of this variable
	*/
	void* Value() const { return m_value; }

	/*
		NumValues
		
		returns how many 'values' this variable has. 
		
		NOTE
		- for example, a vraying variable can have many values, one for each
		grid position ...
	*/
	unsigned NumValues() const { return m_numValues; }

	/*
		SpaceName
		
		returns the name of the 3D space in which this variable has been 
		declared
	*/
	const char* SpaceName() const { return m_spaceName; }

	/*
		Use of this function is prohibited; it is here only
		for the creation of the visibleVars array in the IRIX C shaders
	*/
	void operator=(const DlVar& i_var)
	{
        m_shaderToken = i_var.m_shaderToken;
        m_valueFlags = i_var.m_valueFlags;
        
        m_numValues = 0;
        m_value = 0;

        if( i_var.m_valueFlags & e_ownValue )
        {
            CopyValue(
				i_var.m_spaceName,
				i_var.m_valueClass,
				i_var.m_type,
				i_var.m_numValues,
				i_var.m_value );
        }
        else
        {
            SetValue(
				i_var.m_spaceName,
				i_var.m_valueClass, i_var.m_type,
				i_var.m_numValues,
				i_var.m_value);
        }
	}

    void UseDefaultValue( bool i_useDefault )
    {
        if( i_useDefault )
        {
            m_valueFlags |= e_useDefaultValue;
        }
        else
        {
            m_valueFlags &= (~e_useDefaultValue);
        }
    }

    bool UseDefaultValue( void ) const
	{
		return (m_valueFlags & e_useDefaultValue) != 0;
	}

	/*
		SetType
	*/
	void SetType( const DlTokenInfo::DlType i_type )
	{
		m_type = i_type;
	}

	/*
		Type
	*/
	DlTokenInfo::DlType Type( void ) const
	{
		return m_type;
	}
	
	/*
		ToVarying

		Convert this variable to a vraying one
	*/
	void ToVarying(
		const DlShapeMesh *, const void *i_patch,
		unsigned i_xRes, unsigned i_yRes, DlVar& o_var ) const;

	/*
		public ToVarying (2)

		Converts this variable to a varying variable

		PARAMETERS

			(i_surface, i_patch)
				[in] surface patch identification;

			(i_u, i_v, i_du, i_dv)
				[in] rectangular region to be evaluated;

			o_var
				[out] result will be stored there
	*/
	void ToVarying(
		const DlShapeMesh *i_surface,
		const void *i_patch,
		float i_u, float i_v, float i_du, float i_dv,
		DlVar& o_var ) const;

	/*
		ToUniform

		Converts this variable to an uniform one
	*/
	void ToUniform(
		const DlShapeMesh *, const void *i_patch,
		DlVar& o_var ) const;

	/*
		Size

		returns the size, in bytes, of thie variable
	*/
	unsigned Size( void ) const;

	/*
		NumChannles

		How many channles per single value in this varibale?

		NOTE
		ie : RtVector = 3, RtColor = 3, RtFloat = 1, RtMatrix = 16 ...
	*/	
	unsigned NumChannels( void ) const;

protected:
    void FreeValue( void );

	/* token describing the variable (as declared in the shader) */
	const DlShaderToken *m_shaderToken;

	/* class of the value (e_bad if variable has no value). Note that the class
	   as defined can be different from the one in m_shaderToken. And example
	   is a 'vertex' variable that becomes a 'varying' one in the shader... */
	DlTokenInfo::DlClass m_valueClass;

	/* the type of this variable as declared by user. This may differ from the
	   type inside m_shaderToken. An example is an 'hpoint' variable
	   declared inside a RIB that will become a 'point' variable in a shader. */
	DlTokenInfo::DlType m_type;

    unsigned short m_valueFlags;

	/* */
	unsigned m_numValues;

	/* ptr to the value(s) of this variable; in the case of an RtString,
	   a pointer to the string (char**) */
    void* m_value;				

	/* pointer to space name in which lives the value. NOTE: the DlVar class
	   does not store a copy of the string */
    const char* m_spaceName;

private:

	/* used by DlEvalIntShader */
	void InitWithShaderToken( const DlShaderToken* i_token );
};

#endif
